import inspect
from functools import wraps

from .exceptions import PyCondition, PyConditionError

from .stage import Stage

def wrap_method( func ):
  stage = Stage()
  """
    Wraps a method so that each invariant is called with the 
    self argument.
  """
  @wraps( func )
  def wrapper( *args, **kwargs ):
    ret = func( *args, **kwargs )

    if( stage.name == "Dev" ):
      for invariant in args[0].__invariant__:
        invariant.assertInvariant( args[0] )

    return ret

  wrapper.is_wrapped = True

  return wrapper

class InvariantMeta( type, object ):
  """
    Used to wrap each method in a class.
    It returns the newly created class to be created 
    before it is added to a class as an invariant.
  """
  def __new__(cls, name, bases, attrs):
    for a in attrs:
      if( a != "__metaclass__" and hasattr( attrs[a], "__call__" ) ):
        attrs[a] = wrap_method( attrs[a] )

    return super( InvariantMeta, cls).__new__( cls, name, bases, attrs)

class Invariant( object ):
  """
    This creates a new class that is passed through the InvariantMeta class.
  """
  def __init__( self ):
    self.condition = lambda a: True

  def __call__( self, klass ):
    """
      Ensures that the invriant is applied to only classes.
      Sets the name of the invariant to the class, can be override.
      Add's this invariant to the __invariant__ variable.
    """
    if( not inspect.isclass( klass ) ):
      raise PyConditionError( "Can only put invariant's on classes." )
    self.name = "%s.%s" % ( klass.__module__, klass.__name__ )
    dct = dict( klass.__dict__ )

    if( not hasattr( klass, "__invariant__" ) ):
      dct["__invariant__"] = []
    dct["__invariant__"].append( self )

    t = InvariantMeta( klass.__name__, klass.__bases__, dct )

    return t

class NoopInvariant( Invariant ):
  """
    Does nothing, just returns true.
  """
  def __init__( self ):
    super( NoopInvariant, self ).__init__()

    self.condition = lambda a: True

  def assertInvariant( self, s ):
    return True

class FieldsNotNone( Invariant ):
  """
    Given the fields passed into this invriant they are ensured to be not None
    after each method is called.
  """
  def __init__( self, fields ):
    super( FieldsNotNone, self ).__init__()

    self.fields = fields
 
  def assertInvariant( self, s ):
    for f in self.fields:
      if( getattr( s, f ) == None ):
        raise PyCondition( "Field \"%s\" was None when it should not have been in invariant \"%s\"" % ( f, self.name ) )

class CustomInvariant( Invariant ):
  """
    Allows for a custom method to be passed. The method that is passed must take
    one parameter.
  """
  def __init__( self, name, condition ):
    super( CustomInvariant, self ).__init__()

    self.name = name
    self.condition = condition

  def assertInvariant( self, s ):
    if( not self.condition( s ) ):
      raise PyCondition( "The Custom invariant for \"%s\" did not hold." % self.name )




import unittest

from ..pre import Custom as PreCustom
from ..post import Custom as PostCustom
from ..exceptions import PyCondition

from pyconditions.stage import Stage

class TestStage( unittest.TestCase ):

  def test_both_pre_and_post( self ):
    @PreCustom( "a", lambda a: a[0] == 2 )
    @PostCustom( lambda a: a % 2 == 0 )
    def test( a ):
      return sum( a )

    self.assertRaises( PyCondition, test, [ 1 ] )
    self.assertRaises( PyCondition, test, [ 2, 3 ] )
    self.assertEquals( 2, test( [ 2 ] ) )

    stage = Stage()
    stage.prod()

    self.assertEquals( 5, test( [ 2, 3 ] ) )
    self.assertEquals( 2, test( [ 2 ] ) )




class Singleton(type):
  _instances = {}
  def __call__(cls, *args, **kwargs):
    if cls not in cls._instances:
      cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
    return cls._instances[cls]

_STAGE = Singleton( "_STAGE", ( object, ), {} )

class Stage( _STAGE ):
  def __init__( self ):
    self._name = "Dev"

  @property
  def name( self ):
    return self._name

  def prod( self ):
    self._name = "Prod"

  def dev( self ):
    self._name = "Dev"





from functools import wraps
from .exceptions import PyCondition, PyConditionError
from .inspector import Gadget

from .stage import Stage

class PostCondition( object ):
  def __call__( self, func ):
    stage = Stage()
    self.name = "%s.%s" % ( func.__module__, func.__name__ )

    gadget = Gadget( func )
    doc = gadget.gogoDoc()
    if( not doc and stage.name == "Dev" ):
      doc = ""

    doc += "%s - %s\n" % ( self.name, self.__class__.__name__ )
    func.__doc__ = doc
    @wraps( func )
    def wrapper( *args, **kwargs ):
      returnValue = func( *args, **kwargs )
      if( stage.name == "Dev" ):
        self.assertCondition( returnValue )
      return returnValue
    wrapper._original_func = func

    return wrapper

  def assertCondition( self, returnValue ):
    pass

class NotNone( PostCondition ):
  def assertCondition( self, returnValue ):
    if( returnValue == None ):
      raise PyCondition( "The return value for %s was None" % self.name )

class Custom( PostCondition ):
  def __init__( self, custom ):
    self.custom = custom

  def assertCondition( self, returnValue ):
    if( not self.custom( returnValue ) ):
      raise PyCondition( "The custom condition did not hold for %s" % self.name )

class NotEmpty( PostCondition ):
  def assertCondition( self, returnValue ):
    if( len( returnValue ) == 0 ):
      raise PyCondition( "The return value was empty, meaning the length was 0 for %s" % self.name )



import unittest

from ..inspector import Gadget

def f( test, test2 ):
  pass

class TestGadget( unittest.TestCase ):

  def test_Gadget_retreives_the_parameter_names( self ):

    gadget = Gadget( f )

    self.assertEquals( ( "test", "test2", ), gadget.gogoNames() )

  def test_Gadget_maps_the_name_to_its_index( self ):
    gadget = Gadget( f )

    self.assertEquals( { "test": 0, "test2": 1 }, gadget.gogoMap() )



import unittest

from ..pre import *
from ..exceptions import PyCondition

class TestPre( unittest.TestCase ):

  def test_NotNone_works_with_proper_value( self ):

    @NotNone( "a" )
    def test( a ):
      return a

    self.assertEquals( 1, test( 1 ) )

  def test_NotNone_works_with_multiple_arguements( self ):

    @NotNone( "a" )
    @NotNone( "b" )
    def test( a, b ):
      return a + b

    self.assertEquals( 2, test( 1, 1 ) )

  def test_NotNone_raises( self ):

    @NotNone( "a" )
    def test( a ):
      pass

    self.assertRaises( PyCondition, test, None )

  def test_Between_works_correctly( self ):


    @Between( "a", 0, 10 )
    def test( a, b ):
      pass

    self.assertRaises( PyCondition, test, -1 )
    self.assertRaises( PyCondition, test, 11 )

  def test_Between_multiple_decorators( self ):


    @Between( "a", "a", "z" )
    @Between( "b", 0, 5 )
    def test( b, a ):
      return "%s%d" % ( a, b )

    self.assertEquals( "n2", test( 2, "n" ) )

  def test_GreaterThan_works_correctly( self ):


    @GreaterThan( "b", 0 )
    def test( a, b ):
      return a / b

    self.assertRaises( PyCondition, test, 1, 0 ) 
    self.assertEquals( 1, test( 1, 1 ) )

  def test_LessThan_works_correctly( self ):

    @LessThan( "b", 0 )
    def test( a, b ):
      return a / b

    self.assertRaises( PyCondition, test, 1, 0 ) 
    self.assertEquals( -1, test( 1, -1 ) )

  def test_Custom_works( self ):


    @Custom( "a", lambda a: 0 <= a <= 10 )
    def test( a ):
      return a

    self.assertRaises( PyCondition, test, 11 )
    self.assertEquals( 5, test( 5 ) )

  def test_Instance_works( self ):


    @Instance( "a", int )
    def test( a ):
      return a

    self.assertRaises( PyCondition, test, "a" )
    self.assertEquals( 10, test( 10 ) )

  def test_decorators_work_on_classes( self ):
    class T:

      @NotNone( "a" )
      def __init__( self, a ):
        self.a = a

      @NotNone( "a" )
      def t( self, a ):
        return self.a * a

    t = T( 1 )

    t.t( 10 )

    self.assertRaises( PyCondition, T, None )
    self.assertRaises( PyCondition, t.t, None )
    self.assertEquals( 10, t.t( 10 ) )




import unittest

from ..exceptions import PyCondition, PyConditionError
from ..invariant import *

class TestInvariant( unittest.TestCase ):

  def test_Invariant_properly_sets_metaclass_and_invariant_list( self ):
    @NoopInvariant()
    @NoopInvariant()
    class Test( object ):
      pass

    self.assertTrue( len( Test.__invariant__ ) == 2 )

  def test_Invariant_properly_wraps_each_method( self ):
    @NoopInvariant()
    class Test( object ):
      def test( self ):
        pass

      def test2( self ):
        pass

      def testing( self, a, b ):
        pass

    t = Test()
  
    self.assertTrue( t.test.is_wrapped )
    self.assertTrue( t.test2.is_wrapped )
    self.assertTrue( t.testing.is_wrapped )

  def test_FieldsNotNone( self ):
    @FieldsNotNone( [ "test", "test2" ] )
    class Test( object ):
      def __init__( self ):
        self.test = 1
        self.test2 = 2

      def test3( self ):
        pass

    t = Test()
    t.test = None

    self.assertRaises( PyCondition, t.test3 )

    t.test = 1
    t.test2 = None

    self.assertRaises( PyCondition, t.test3 )

  def test_CustomInvariant( self ):
    def invariant( self ):
      return self.test == 1

    @CustomInvariant( "test", invariant )
    class Test( object ):
      def __init__( self ):
        self.test = 1

      def test2( self ):
        self.test = 2

    t = Test()

    self.assertRaises( PyCondition, t.test2 )

  def test_Invariant_cannot_be_applied_to_non_classes( self ):
    def test():
      @FieldsNotNone( [ "test" ] )
      def testing():
        pass

    self.assertRaises( PyConditionError, test )



import sys

rr = None
try:
  xrange( 0 )
  rr = xrange
except:
  rr = range

class Gadget:
  """
    A Gadget is a utility class that is used to gather the information
    to perform the conditional checks on specific parameters.
  """

  def __init__( self, func ):
    self.func = func

    self._varnames = None
    self._argcounts = None
    self._map = None
    self.name = func.__name__
    self.module = func.__module__

  def gogoNames( self ):
    """
      This will return the names of the parameters. But, because in 
      python > 3.* the attribute of the function changed from
      func_code to __code__ the version check is required.
    """
    if( not self._varnames ):
      if( sys.hexversion >= 0x03000000 ):
        self._varnames = self.func.__code__.co_varnames
      else:
        self._varnames = self.func.func_code.co_varnames

    return self._varnames

  def gogoIndexes( self ):
    """
      This will return the indexes of the parameters which can be used 
      to get the values themselves. The system version needs to be checked
      because of the change from func_code to __code__.
    """
    if( not self._argcounts ):
      if( sys.hexversion >= 0x03000000 ):
        self._argcounts = rr( self.func.__code__.co_argcount )
      else:
        self._argcounts = rr( self.func.func_code.co_argcount )

    return self._argcounts

  def gogoMap( self ):
    """
      Used to build up a dict that maps parameter name to 
      it's associated index.
    """
    if( not self._map ):
      names = self.gogoNames()
      indexes = self.gogoIndexes()

      ret = {}

      for i in indexes:
        ret[names[i]] = i

      self._map = ret
    
    return self._map

  def gogoDoc( self ):
    return self.func.__doc__



from distutils.core import setup

setup(
    name='pyConditions',
    version='0.5.1',
    author='Sean Reed',
    author_email='streed@mail.roanoke.edu',
    packages=['pyconditions', 'pyconditions.test'],
    url='http://github.com/streed/pyConditions',
    license='LICENSE.txt',
    description='Commenting sucks so let your code do it for you with Guava like preconditions that also actually do something.',
    long_description=open('README.rst').read(),
    classifiers=[
      "Development Status :: 4 - Beta",
      "License :: OSI Approved :: Apache Software License",
      "Programming Language :: Python :: 2.6",
      "Programming Language :: Python :: 2.7",
      "Programming Language :: Python :: 3.2",
      "Programming Language :: Python :: 3.3",
      "Programming Language :: Python :: Implementation :: CPython",
      "Programming Language :: Python :: Implementation :: PyPy",
      "Topic :: Software Development :: Quality Assurance"
    ],
)



from functools import wraps
from .inspector import  Gadget
from .exceptions import PyCondition, PyConditionError

from .stage import Stage

class Pre( object ):
  def __init__( self ):
    self.funcTable = {}

class PreCondition( object ):
  context = Pre()
  def __init__( self, name ):
    self.name = name

  def _map( self, func ):
    try:
      func = func._original_func
      func.__doc__ = func._original_func.__doc__
    except:
      pass

    gadget = Gadget( func )

    m = gadget.gogoMap()
    params = [ v[1] for v in sorted( [ ( i, vv ) for vv, i in m.items() ], key=lambda a: a[0] ) ]
    name = ( func.__module__, func.__name__, "( %s )" % ",".join( params ) )
    self.funcName = "%s.%s%s" % name

    if( name in self.context.funcTable ):
      self.argMap = PreCondition.context.funcTable[name]
    else:
      gadget = Gadget( func )
      self.argMap = gadget.gogoMap()

      PreCondition.context.funcTable[name] = self.argMap

  def assertCondition( self, *args, **kwargs ):
    pass

  def __call__( self, func ):
    wrapper = func
    stage = Stage()

    if( not func.__doc__ and stage.name == "Dev" ):
      doc = ""
      func.__doc__ = ""

    func.__doc__ += "  %s - %s\n" % ( self.name, self.__class__.__name__ )

    self._map( func )
    
    @wraps( func )
    def wrapper( *args, **kwargs ):
      if( stage.name == "Dev" ):
        self.assertCondition( *args, **kwargs )
      return func( *args, **kwargs )

    wrapper._original_func = func

    return wrapper

class NotNone( PreCondition ):

  def __init__( self, name ):
    super( NotNone, self ).__init__( name )
    self.name = name

  def assertCondition( self, *args, **kwargs ):
      if( args[self.argMap[self.name]] == None ):
        raise PyCondition( "%s is None in function %s" % ( self.name, self.funcName ) )

class Between( PreCondition ):
  def __init__( self, name, lower, upper ):
    super( Between, self ).__init__(  name )
    self.lower = lower
    self.upper = upper

  def assertCondition( self, *args, **kwargs ):
    v = args[self.argMap[self.name]]

    if( not ( self.lower <= v <= self.upper ) ):
      raise PyCondition( "%s <= %s <= %s did not hold for parameter %s in function %s" % ( self.lower, v, self.upper, self.name, self.funcName ) )

class GreaterThan( PreCondition ):
  def __init__( self, name, lower ):
    super( GreaterThan, self ).__init__( name )
    self.lower = lower

  def assertCondition( self, *args, **kwargs ):
    v = args[self.argMap[self.name]]

    if( not ( self.lower < v ) ):
      raise PyCondition( "%s > %s did not hold for parameter %s in function %s" % ( v, self.lower, self.name, self.funcName ) )

class LessThan( PreCondition ):
  def __init__( self, name, upper ):
    super( LessThan, self ).__init__( name )
    self.upper = upper

  def assertCondition( self, *args, **kwargs ):
    v = args[self.argMap[self.name]]

    if( not ( self.upper > v ) ):
      raise PyCondition( "%s < %s did not hold for parameter %s in function %s" % ( v, self.upper, self.name, self.funcName ) )

class Custom( PreCondition ):
  def __init__( self, name, check ):
    super( Custom, self ).__init__( name ) 
    self.check = check

  def assertCondition( self, *args, **kwargs ):
    v = args[self.argMap[self.name]]

    if( not self.check( v ) ):
      raise PyCondition( "%s did not hold for the parameter %s in the custom condition in function  %s" % ( v, self.name, self.funcName ) )

class Instance( PreCondition ):
  def __init__( self, name, klass ):
    super( Instance, self ).__init__( name )
    self.klass = klass

  def assertCondition( self, *args, **kwargs ):
    v = args[self.argMap[self.name]]

    if( not isinstance( v, self.klass ) ):
      raise PyCondition( "%s is not a %s for parameter %s in function %s" % ( v, self.klass, self.name, self.funcName ) )




import unittest

from ..post import *
from ..exceptions import PyCondition

class TestPost( unittest.TestCase ):

  def test_NotNone( self ):

    @NotNone()
    def test():
      return None

    self.assertRaises( PyCondition, test )

  def test_Custom( self ):

    @Custom( lambda a: a % 2 == 0 )
    def even( a ):
      return a

    self.assertRaises( PyCondition, even, 3 )
    self.assertEquals( 2, even( 2 ) )

  def test_NotEmpty( self ):

    @NotEmpty()
    def empty( a ):
      return a

    self.assertRaises( PyCondition, empty, [] )
    self.assertEquals( [ 1 ], empty( [ 1 ] ) )

  def test_decorators_stacked( self ):
    
    @Custom( lambda a: a[0] == 2 )
    @NotEmpty()
    def stacked( a ):
      return a

    self.assertRaises( PyCondition, stacked, [] )
    self.assertRaises( PyCondition, stacked, [ 3 ] )
    self.assertEquals( [ 2 ], stacked( [ 2 ] ) )
